/*
 * Copyright (c) 2015 Nicholas Fraser
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#define MPACK_INTERNAL 1

#include "mpack-reader.h"

#if MPACK_READER

void mpack_reader_init(mpack_reader_t* reader, char* buffer, size_t size, size_t count) {
    mpack_memset(reader, 0, sizeof(*reader));
    reader->buffer = buffer;
    reader->size = size;
    reader->left = count;
    MPACK_READER_TRACK(reader, mpack_track_init(&reader->track));
}

void mpack_reader_init_error(mpack_reader_t* reader, mpack_error_t error) {
    mpack_memset(reader, 0, sizeof(*reader));
    reader->error = error;
}

void mpack_reader_init_data(mpack_reader_t* reader, const char* data, size_t count) {
    mpack_memset(reader, 0, sizeof(*reader));
    reader->left = count;

    // unfortunately we have to cast away the const to store the buffer,
    // but we won't be modifying it because there's no fill function.
    // the buffer size is left at 0 to ensure no fill function can be
    // set or used (see mpack_reader_set_fill().)
    #ifdef __cplusplus
    reader->buffer = const_cast<char*>(data);
    #else
    reader->buffer = (char*)data;
    #endif

    MPACK_READER_TRACK(reader, mpack_track_init(&reader->track));
}

#if MPACK_STDIO
typedef struct mpack_file_reader_t {
    FILE* file;
    char buffer[MPACK_BUFFER_SIZE];
} mpack_file_reader_t;

static size_t mpack_file_reader_fill(mpack_reader_t* reader, char* buffer, size_t count) {
    mpack_file_reader_t* file_reader = (mpack_file_reader_t*)reader->context;
    return fread((void*)buffer, 1, count, file_reader->file);
}

static void mpack_file_reader_teardown(mpack_reader_t* reader) {
    mpack_file_reader_t* file_reader = (mpack_file_reader_t*)reader->context;

    if (file_reader->file) {
        int ret = fclose(file_reader->file);
        file_reader->file = NULL;
        if (ret != 0)
            mpack_reader_flag_error(reader, mpack_error_io);
    }

    MPACK_FREE(file_reader);
}

void mpack_reader_init_file(mpack_reader_t* reader, const char* filename) {
    mpack_file_reader_t* file_reader = (mpack_file_reader_t*) MPACK_MALLOC(sizeof(mpack_file_reader_t));
    if (file_reader == NULL) {
        mpack_reader_init_error(reader, mpack_error_memory);
        return;
    }

    file_reader->file = fopen(filename, "rb");
    if (file_reader->file == NULL) {
        mpack_reader_init_error(reader, mpack_error_io);
        MPACK_FREE(file_reader);
        return;
    }

    mpack_reader_init(reader, file_reader->buffer, sizeof(file_reader->buffer), 0);
    mpack_reader_set_context(reader, file_reader);
    mpack_reader_set_fill(reader, mpack_file_reader_fill);
    mpack_reader_set_teardown(reader, mpack_file_reader_teardown);
}
#endif

mpack_error_t mpack_reader_destroy_impl(mpack_reader_t* reader, bool cancel) {
    MPACK_UNUSED(cancel);
    MPACK_READER_TRACK(reader, mpack_track_destroy(&reader->track, cancel));

    if (reader->teardown)
        reader->teardown(reader);
    reader->teardown = NULL;

    #if MPACK_SETJMP
    if (reader->jump_env)
        MPACK_FREE(reader->jump_env);
    reader->jump_env = NULL;
    #endif

    return reader->error;
}

void mpack_reader_destroy_cancel(mpack_reader_t* reader) {
    mpack_reader_destroy_impl(reader, true);
}

mpack_error_t mpack_reader_destroy(mpack_reader_t* reader) {
    return mpack_reader_destroy_impl(reader, false);
}

size_t mpack_reader_remaining(mpack_reader_t* reader, const char** data) {
    MPACK_READER_TRACK(reader, mpack_track_check_empty(&reader->track));
    if (data)
        *data = reader->buffer + reader->pos;
    return reader->left;
}

void mpack_reader_flag_error(mpack_reader_t* reader, mpack_error_t error) {
    mpack_log("reader %p setting error %i: %s\n", reader, (int)error, mpack_error_to_string(error));

    if (reader->error == mpack_ok) {
        reader->error = error;
        #if MPACK_SETJMP
        if (reader->jump_env)
            longjmp(*reader->jump_env, 1);
        #endif
    }
}

// A helper to call the reader fill function. This makes sure it's
// implemented and guards against overflow in case it returns -1.
static inline size_t mpack_fill(mpack_reader_t* reader, char* p, size_t count) {
    if (!reader->fill)
        return 0;
    size_t ret = reader->fill(reader, p, count);
    if (ret == ((size_t)(-1)))
        return 0;
    return ret;
}

// Reads count bytes into p. Used when there are not enough bytes
// left in the buffer to satisfy a read.
void mpack_read_native_big(mpack_reader_t* reader, char* p, size_t count) {
    if (mpack_reader_error(reader) != mpack_ok) {
        mpack_memset(p, 0, count);
        return;
    }

    mpack_log("big read for %i bytes into %p, %i left in buffer, buffer size %i\n",
            (int)count, p, (int)reader->left, (int)reader->size);

    if (count <= reader->left) {
        mpack_assert(0,
                "big read requested for %i bytes, but there are %i bytes "
                "left in buffer. call mpack_read_native() instead",
                (int)count, (int)reader->left);
        mpack_reader_flag_error(reader, mpack_error_bug);
        mpack_memset(p, 0, count);
        return;
    }

    if (reader->size == 0) {
        // somewhat debatable what error should be returned here. when
        // initializing a reader with an in-memory buffer it's not
        // necessarily a bug if the data is blank; it might just have
        // been truncated to zero. for this reason we return the same
        // error as if the data was truncated.
        mpack_reader_flag_error(reader, mpack_error_io);
        mpack_memset(p, 0, count);
        return;
    }

    // flush what's left of the buffer
    if (reader->left > 0) {
        mpack_log("flushing %i bytes remaining in buffer\n", (int)reader->left);
        mpack_memcpy(p, reader->buffer + reader->pos, reader->left);
        count -= reader->left;
        p += reader->left;
        reader->pos += reader->left;
        reader->left = 0;
    }

    // we read only in multiples of the buffer size. read the middle portion, if any
    size_t middle = count - (count % reader->size);
    if (middle > 0) {
        mpack_log("reading %i bytes in middle\n", (int)middle);
        if (mpack_fill(reader, p, middle) < middle) {
            mpack_reader_flag_error(reader, mpack_error_io);
            mpack_memset(p, 0, count);
            return;
        }
        count -= middle;
        p += middle;
        if (count == 0)
            return;
    }

    // fill the buffer
    reader->pos = 0;
    reader->left = mpack_fill(reader, reader->buffer, reader->size);
    mpack_log("filled %i bytes into buffer\n", (int)reader->left);
    if (reader->left < count) {
        mpack_reader_flag_error(reader, mpack_error_io);
        mpack_memset(p, 0, count);
        return;
    }

    // serve the remainder
    mpack_log("serving %i remaining bytes from %p to %p\n", (int)count, reader->buffer+reader->pos,p);
    mpack_memcpy(p, reader->buffer + reader->pos, count);
    reader->pos += count;
    reader->left -= count;
}

void mpack_skip_bytes(mpack_reader_t* reader, size_t count) {
    // TODO: This is currently very slow, potentially even slower than just
    // reading the data. Skip needs to be implemented properly.
    char c[128];
    size_t i = 0;
    while (i < count && mpack_reader_error(reader) == mpack_ok) {
        size_t amount = ((count - i) > sizeof(c)) ? sizeof(c) : (count - i);
        mpack_read_bytes(reader, c, amount);
        i += amount;
    }
}

void mpack_read_bytes(mpack_reader_t* reader, char* p, size_t count) {
    mpack_reader_track_bytes(reader, count);
    mpack_read_native(reader, p, count);
}

// internal inplace reader for when it straddles the end of the buffer
// this is split out to inline the common case, although this isn't done
// right now because we can't inline tracking yet
static const char* mpack_read_bytes_inplace_big(mpack_reader_t* reader, size_t count) {

    // we should only arrive here from inplace straddle; this should already be checked
    mpack_assert(mpack_reader_error(reader) == mpack_ok, "already in error state? %s",
            mpack_error_to_string(mpack_reader_error(reader)));
    mpack_assert(reader->left < count, "already enough bytes in buffer: %i left, %i count", (int)reader->left, (int)count);

    // we'll need a fill function to get more data
    if (!reader->fill) {
        mpack_reader_flag_error(reader, mpack_error_io);
        return NULL;
    }

    // make sure the buffer is big enough to actually fit the data
    if (count > reader->size) {
        mpack_reader_flag_error(reader, mpack_error_too_big);
        return NULL;
    }

    // shift the remaining data back to the start and fill the buffer back up
    mpack_memmove(reader->buffer, reader->buffer + reader->pos, reader->left);
    reader->pos = 0;
    reader->left += mpack_fill(reader, reader->buffer + reader->left, reader->size - reader->left);
    if (reader->left < count) {
        mpack_reader_flag_error(reader, mpack_error_io);
        return NULL;
    }
    reader->pos += count;
    reader->left -= count;
    return reader->buffer;
}

const char* mpack_read_bytes_inplace(mpack_reader_t* reader, size_t count) {
    if (mpack_reader_error(reader) != mpack_ok)
        return NULL;

    mpack_reader_track_bytes(reader, count);

    // if we have enough bytes already in the buffer, we can return it directly.
    if (reader->left >= count) {
        reader->pos += count;
        reader->left -= count;
        return reader->buffer + reader->pos - count;
    }

    return mpack_read_bytes_inplace_big(reader, count);
}

mpack_tag_t mpack_read_tag(mpack_reader_t* reader) {
    mpack_tag_t var;
    mpack_memset(&var, 0, sizeof(var));
    var.type = mpack_type_nil;

    // get the type
    uint8_t type = mpack_read_native_u8(reader);
    if (mpack_reader_error(reader))
        return var;
    mpack_reader_track_element(reader);

    // unfortunately, by far the fastest way to parse a tag is to switch
    // on the first byte, and to explicitly list every possible byte. so for
    // infix types, the list of cases is quite large. the compiler optimizes
    // this nicely (and it takes very little space.)
    switch (type) {

        // positive fixnum
        case 0x00: case 0x01: case 0x02: case 0x03: case 0x04: case 0x05: case 0x06: case 0x07:
        case 0x08: case 0x09: case 0x0a: case 0x0b: case 0x0c: case 0x0d: case 0x0e: case 0x0f:
        case 0x10: case 0x11: case 0x12: case 0x13: case 0x14: case 0x15: case 0x16: case 0x17:
        case 0x18: case 0x19: case 0x1a: case 0x1b: case 0x1c: case 0x1d: case 0x1e: case 0x1f:
        case 0x20: case 0x21: case 0x22: case 0x23: case 0x24: case 0x25: case 0x26: case 0x27:
        case 0x28: case 0x29: case 0x2a: case 0x2b: case 0x2c: case 0x2d: case 0x2e: case 0x2f:
        case 0x30: case 0x31: case 0x32: case 0x33: case 0x34: case 0x35: case 0x36: case 0x37:
        case 0x38: case 0x39: case 0x3a: case 0x3b: case 0x3c: case 0x3d: case 0x3e: case 0x3f:
        case 0x40: case 0x41: case 0x42: case 0x43: case 0x44: case 0x45: case 0x46: case 0x47:
        case 0x48: case 0x49: case 0x4a: case 0x4b: case 0x4c: case 0x4d: case 0x4e: case 0x4f:
        case 0x50: case 0x51: case 0x52: case 0x53: case 0x54: case 0x55: case 0x56: case 0x57:
        case 0x58: case 0x59: case 0x5a: case 0x5b: case 0x5c: case 0x5d: case 0x5e: case 0x5f:
        case 0x60: case 0x61: case 0x62: case 0x63: case 0x64: case 0x65: case 0x66: case 0x67:
        case 0x68: case 0x69: case 0x6a: case 0x6b: case 0x6c: case 0x6d: case 0x6e: case 0x6f:
        case 0x70: case 0x71: case 0x72: case 0x73: case 0x74: case 0x75: case 0x76: case 0x77:
        case 0x78: case 0x79: case 0x7a: case 0x7b: case 0x7c: case 0x7d: case 0x7e: case 0x7f:
            var.type = mpack_type_uint;
            var.v.u = type;
            return var;

        // negative fixnum
        case 0xe0: case 0xe1: case 0xe2: case 0xe3: case 0xe4: case 0xe5: case 0xe6: case 0xe7:
        case 0xe8: case 0xe9: case 0xea: case 0xeb: case 0xec: case 0xed: case 0xee: case 0xef:
        case 0xf0: case 0xf1: case 0xf2: case 0xf3: case 0xf4: case 0xf5: case 0xf6: case 0xf7:
        case 0xf8: case 0xf9: case 0xfa: case 0xfb: case 0xfc: case 0xfd: case 0xfe: case 0xff:
            var.type = mpack_type_int;
            var.v.i = (int8_t)type;
            return var;

        // fixmap
        case 0x80: case 0x81: case 0x82: case 0x83: case 0x84: case 0x85: case 0x86: case 0x87:
        case 0x88: case 0x89: case 0x8a: case 0x8b: case 0x8c: case 0x8d: case 0x8e: case 0x8f:
            var.type = mpack_type_map;
            var.v.n = type & ~0xf0;
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_map, var.v.n));
            return var;

        // fixarray
        case 0x90: case 0x91: case 0x92: case 0x93: case 0x94: case 0x95: case 0x96: case 0x97:
        case 0x98: case 0x99: case 0x9a: case 0x9b: case 0x9c: case 0x9d: case 0x9e: case 0x9f:
            var.type = mpack_type_array;
            var.v.n = type & ~0xf0;
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_array, var.v.n));
            return var;

        // fixstr
        case 0xa0: case 0xa1: case 0xa2: case 0xa3: case 0xa4: case 0xa5: case 0xa6: case 0xa7:
        case 0xa8: case 0xa9: case 0xaa: case 0xab: case 0xac: case 0xad: case 0xae: case 0xaf:
        case 0xb0: case 0xb1: case 0xb2: case 0xb3: case 0xb4: case 0xb5: case 0xb6: case 0xb7:
        case 0xb8: case 0xb9: case 0xba: case 0xbb: case 0xbc: case 0xbd: case 0xbe: case 0xbf:
            var.type = mpack_type_str;
            var.v.l = type & ~0xe0;
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_str, var.v.l));
            return var;

        // nil
        case 0xc0:
            var.type = mpack_type_nil;
            return var;

        // bool
        case 0xc2: case 0xc3:
            var.type = mpack_type_bool;
            var.v.b = type & 1;
            return var;

        // bin8
        case 0xc4:
            var.type = mpack_type_bin;
            var.v.l = mpack_read_native_u8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_bin, var.v.l));
            return var;

        // bin16
        case 0xc5:
            var.type = mpack_type_bin;
            var.v.l = mpack_read_native_u16(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_bin, var.v.l));
            return var;

        // bin32
        case 0xc6:
            var.type = mpack_type_bin;
            var.v.l = mpack_read_native_u32(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_bin, var.v.l));
            return var;

        // ext8
        case 0xc7:
            var.type = mpack_type_ext;
            var.v.l = mpack_read_native_u8(reader);
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // ext16
        case 0xc8:
            var.type = mpack_type_ext;
            var.v.l = mpack_read_native_u16(reader);
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // ext32
        case 0xc9:
            var.type = mpack_type_ext;
            var.v.l = mpack_read_native_u32(reader);
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // float
        case 0xca:
            var.type = mpack_type_float;
            var.v.f = mpack_read_native_float(reader);
            return var;

        // double
        case 0xcb:
            var.type = mpack_type_double;
            var.v.d = mpack_read_native_double(reader);
            return var;

        // uint8
        case 0xcc:
            var.type = mpack_type_uint;
            var.v.u = mpack_read_native_u8(reader);
            return var;

        // uint16
        case 0xcd:
            var.type = mpack_type_uint;
            var.v.u = mpack_read_native_u16(reader);
            return var;

        // uint32
        case 0xce:
            var.type = mpack_type_uint;
            var.v.u = mpack_read_native_u32(reader);
            return var;

        // uint64
        case 0xcf:
            var.type = mpack_type_uint;
            var.v.u = mpack_read_native_u64(reader);
            return var;

        // int8
        case 0xd0:
            var.type = mpack_type_int;
            var.v.i = mpack_read_native_i8(reader);
            return var;

        // int16
        case 0xd1:
            var.type = mpack_type_int;
            var.v.i = mpack_read_native_i16(reader);
            return var;

        // int32
        case 0xd2:
            var.type = mpack_type_int;
            var.v.i = mpack_read_native_i32(reader);
            return var;

        // int64
        case 0xd3:
            var.type = mpack_type_int;
            var.v.i = mpack_read_native_i64(reader);
            return var;

        // fixext1
        case 0xd4:
            var.type = mpack_type_ext;
            var.v.l = 1;
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // fixext2
        case 0xd5:
            var.type = mpack_type_ext;
            var.v.l = 2;
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // fixext4
        case 0xd6:
            var.type = mpack_type_ext;
            var.v.l = 4;
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // fixext8
        case 0xd7:
            var.type = mpack_type_ext;
            var.v.l = 8;
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // fixext16
        case 0xd8:
            var.type = mpack_type_ext;
            var.v.l = 16;
            var.exttype = mpack_read_native_i8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_ext, var.v.l));
            return var;

        // str8
        case 0xd9:
            var.type = mpack_type_str;
            var.v.l = mpack_read_native_u8(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_str, var.v.l));
            return var;

        // str16
        case 0xda:
            var.type = mpack_type_str;
            var.v.l = mpack_read_native_u16(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_str, var.v.l));
            return var;

        // str32
        case 0xdb:
            var.type = mpack_type_str;
            var.v.l = mpack_read_native_u32(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_str, var.v.l));
            return var;

        // array16
        case 0xdc:
            var.type = mpack_type_array;
            var.v.n = mpack_read_native_u16(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_array, var.v.n));
            return var;

        // array32
        case 0xdd:
            var.type = mpack_type_array;
            var.v.n = mpack_read_native_u32(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_array, var.v.n));
            return var;

        // map16
        case 0xde:
            var.type = mpack_type_map;
            var.v.n = mpack_read_native_u16(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_map, var.v.n));
            return var;

        // map32
        case 0xdf:
            var.type = mpack_type_map;
            var.v.n = mpack_read_native_u32(reader);
            MPACK_READER_TRACK(reader, mpack_track_push(&reader->track, mpack_type_map, var.v.n));
            return var;

        // reserved
        case 0xc1:
            break;
    }

    // unrecognized type
    mpack_reader_flag_error(reader, mpack_error_invalid);
    return var;
}

void mpack_discard(mpack_reader_t* reader) {
    mpack_tag_t var = mpack_read_tag(reader);
    if (mpack_reader_error(reader))
        return;
    switch (var.type) {
        case mpack_type_str:
            mpack_skip_bytes(reader, var.v.l);
            mpack_done_str(reader);
            break;
        case mpack_type_bin:
            mpack_skip_bytes(reader, var.v.l);
            mpack_done_bin(reader);
            break;
        case mpack_type_ext:
            mpack_skip_bytes(reader, var.v.l);
            mpack_done_ext(reader);
            break;
        case mpack_type_array: {
            for (; var.v.n > 0; --var.v.n) {
                mpack_discard(reader);
                if (mpack_reader_error(reader))
                    break;
            }
            break;
        }
        case mpack_type_map: {
            for (; var.v.n > 0; --var.v.n) {
                mpack_discard(reader);
                mpack_discard(reader);
                if (mpack_reader_error(reader))
                    break;
            }
            break;
        }
        default:
            break;
    }
}

#if MPACK_READ_TRACKING
void mpack_done_array(mpack_reader_t* reader) {
    MPACK_READER_TRACK(reader, mpack_track_pop(&reader->track, mpack_type_array));
}

void mpack_done_map(mpack_reader_t* reader) {
    MPACK_READER_TRACK(reader, mpack_track_pop(&reader->track, mpack_type_map));
}

void mpack_done_str(mpack_reader_t* reader) {
    MPACK_READER_TRACK(reader, mpack_track_pop(&reader->track, mpack_type_str));
}

void mpack_done_bin(mpack_reader_t* reader) {
    MPACK_READER_TRACK(reader, mpack_track_pop(&reader->track, mpack_type_bin));
}

void mpack_done_ext(mpack_reader_t* reader) {
    MPACK_READER_TRACK(reader, mpack_track_pop(&reader->track, mpack_type_ext));
}

void mpack_done_type(mpack_reader_t* reader, mpack_type_t type) {
    MPACK_READER_TRACK(reader, mpack_track_pop(&reader->track, type));
}
#endif

#if MPACK_DEBUG && MPACK_STDIO && MPACK_SETJMP && !MPACK_NO_PRINT
static void mpack_debug_print_element(mpack_reader_t* reader, size_t depth) {
    mpack_tag_t val = mpack_read_tag(reader);
    switch (val.type) {

        case mpack_type_nil:
            printf("null");
            break;
        case mpack_type_bool:
            printf(val.v.b ? "true" : "false");
            break;

        case mpack_type_float:
            printf("%f", val.v.f);
            break;
        case mpack_type_double:
            printf("%f", val.v.d);
            break;

        case mpack_type_int:
            printf("%"PRIi64, val.v.i);
            break;
        case mpack_type_uint:
            printf("%"PRIu64, val.v.u);
            break;

        case mpack_type_bin:
            // skip data
            for (size_t i = 0; i < val.v.l; ++i)
                mpack_read_native_u8(reader);
            printf("<binary data>");
            mpack_done_bin(reader);
            break;

        case mpack_type_ext:
            // skip data
            for (size_t i = 0; i < val.v.l; ++i)
                mpack_read_native_u8(reader);
            printf("<ext data of type %i>", val.exttype);
            mpack_done_ext(reader);
            break;

        case mpack_type_str:
            putchar('"');
            for (size_t i = 0; i < val.v.l; ++i) {
                char c;
                mpack_read_bytes(reader, &c, 1);
                switch (c) {
                    case '\n': printf("\\n"); break;
                    case '\\': printf("\\\\"); break;
                    case '"': printf("\\\""); break;
                    default: putchar(c); break;
                }
            }
            putchar('"');
            mpack_done_str(reader);
            break;

        case mpack_type_array:
            printf("[\n");
            for (size_t i = 0; i < val.v.n; ++i) {
                for (size_t j = 0; j < depth + 1; ++j)
                    printf("    ");
                mpack_debug_print_element(reader, depth + 1);
                if (i != val.v.n - 1)
                    putchar(',');
                putchar('\n');
            }
            for (size_t i = 0; i < depth; ++i)
                printf("    ");
            putchar(']');
            mpack_done_array(reader);
            break;

        case mpack_type_map:
            printf("{\n");
            for (size_t i = 0; i < val.v.n; ++i) {
                for (size_t j = 0; j < depth + 1; ++j)
                    printf("    ");
                mpack_debug_print_element(reader, depth + 1);
                printf(": ");
                mpack_debug_print_element(reader, depth + 1);
                if (i != val.v.n - 1)
                    putchar(',');
                putchar('\n');
            }
            for (size_t i = 0; i < depth; ++i)
                printf("    ");
            putchar('}');
            mpack_done_map(reader);
            break;
    }
}

void mpack_debug_print(const char* data, int len) {
    mpack_reader_t reader;
    mpack_reader_init_data(&reader, data, len);
    if (MPACK_READER_SETJMP(&reader)) {
        printf("<mpack parsing error %s>\n", mpack_error_to_string(mpack_reader_error(&reader)));
        return;
    }

    int depth = 2;
    for (int i = 0; i < depth; ++i)
        printf("    ");
    mpack_debug_print_element(&reader, depth);
    putchar('\n');

    if (mpack_reader_remaining(&reader, NULL) > 0)
        printf("<%i extra bytes at end of mpack>\n", (int)mpack_reader_remaining(&reader, NULL));
}
#endif

#endif

